/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db;

import ch.ehi.basics.logging.EhiLogger;
import ch.ehi.ili2db.base.Ili2db;
import ch.ehi.ili2db.base.Ili2dbException;
import ch.ehi.ili2db.gui.Config;
import ch.ehi.ili2db.gui.AbstractDbPanelDescriptor;
import ch.ehi.ili2db.base.DbUrlConverter;


/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 07.02.2005 $
 */
public abstract class AbstractMain {
	public abstract String getAPP_NAME(); 
	public abstract String getDB_PRODUCT_NAME(); 
	public abstract String getJAR_NAME(); 
	public abstract AbstractDbPanelDescriptor getDbPanelDescriptor();
	protected abstract void doDbOptions();
	protected abstract int doArgs(String args[],int argi,Config config);
	protected void initConfig(Config config)
	{
		config.setSender(getAPP_NAME()+"-"+getVersion());
		config.setModeldir(Ili2db.XTF_DIR+";"+Ili2db.JAR_DIR);
		config.setModels(Ili2db.XTF);
		config.setDefaultSrsAuthority("EPSG");
		config.setDefaultSrsCode("21781");
	}
	protected abstract DbUrlConverter getDbUrlConverter();

	public void domain(String args[]){
		Config config=new Config();
		initConfig(config);
		if(args.length==0){
			ch.ehi.ili2db.gui.MainWizard.main(config,getAPP_HOME(),getAPP_NAME(),getDbPanelDescriptor(),getDbUrlConverter());
			return;
		}
		int argi=0;
		boolean doGui=false;
		for(;argi<args.length;){
			while(argi<args.length){
				int oldargi=argi;
				argi=doArgs(args,argi,config);
				if(argi==oldargi){
					break;
				}
			}
			String arg=args[argi];
			if(arg.equals("-modeldir")){
				argi++;
				config.setModeldir(args[argi]);
				argi++;
			}else if(arg.equals("-models")){
				argi++;
				config.setModels(args[argi]);
				argi++;
			}else if(arg.equals("-gui")){
				doGui=true;
				argi++;
			}else if(arg.equals("-import")){
				config.setFunction(Config.FC_IMPORT);
				argi++;
			}else if(arg.equals("-export")){
				config.setFunction(Config.FC_EXPORT);
				argi++;
			}else if(arg.equals("-schemaimport")){
				config.setFunction(Config.FC_SCHEMAIMPORT);
				argi++;
			}else if(arg.equals("-trace")){
				EhiLogger.getInstance().setTraceFilter(false); 
				argi++;
			}else if(arg.equals("-dropscript")){
				argi++;
				config.setDropscript(args[argi]);
				argi++;
			}else if(arg.equals("-createscript")){
				argi++;
				config.setCreatescript(args[argi]);
				argi++;
			}else if(arg.equals("-mapconfig")){
				argi++;
				config.setMappingConfig(args[argi]);
				argi++;
			}else if(arg.equals("-defaultSrsAuth")){
				argi++;
				String auth=args[argi];
				if(auth.equalsIgnoreCase("NULL")){
					auth=null;
				}
				config.setDefaultSrsAuthority(auth);
				argi++;
			}else if(arg.equals("-defaultSrsCode")){
				argi++;
				config.setDefaultSrsCode(args[argi]);
				argi++;
			}else if(arg.equals("-version")){
				printVersion();
				return;
			}else if(arg.equals("-help")){
					printVersion ();
					System.err.println();
					printDescription ();
					System.err.println();
					printUsage ();
					System.err.println();
					System.err.println("OPTIONS");
					System.err.println();
					System.err.println("-import               do an import.");
					System.err.println("-export               do an export.");
					System.err.println("-schemaimport               do an schema import.");
					doDbOptions();
					System.err.println("-defaultSrsAuth  auth Default SRS authority "+config.getDefaultSrsAuthority());
					System.err.println("-defaultSrsCode  code Default SRS code "+config.getDefaultSrsCode());
					System.err.println("-modeldir  path       Path(s) of directories containing ili-files.");
					System.err.println("-models modelname     Name(s) of ili-models to generate an db schema for.");
					System.err.println("-createscript filename  Generate a sql script that creates the db schema.");
					System.err.println("-dropscript filename  Generate a sql script that drops the generated db schema.");
					System.err.println("-mapconfig filename   Name of config file, that controls the schema mapping.");
					System.err.println("-gui                  start GUI.");
					System.err.println("-trace                enable trace messages.");
					System.err.println("-help                 Display this help text.");
					System.err.println("-version              Display the version of "+getAPP_NAME());
					System.err.println();
					return;
				
			}else if(arg.startsWith("-")){
				EhiLogger.logAdaption(arg+": unknown option; ignored");
				argi++;
			}else{
				break;
			}
		}
		if(argi<args.length){
			config.setXtffile(args[argi]);
		}
		if(doGui){
			ch.ehi.ili2db.gui.MainWizard.main(config,getAPP_HOME(),getAPP_NAME(),getDbPanelDescriptor(),getDbUrlConverter());
		}else{
			config.setDburl(getDbUrlConverter().makeUrl(config));
			try {
				Ili2db.run(config,getAPP_HOME());
			} catch (Ili2dbException ex) {
				EhiLogger.logError(ex);
			}
		}
		
	}

	private void printVersion ()
	{
	  System.err.println("INTERLIS 2-loader for "+getDB_PRODUCT_NAME()+", Version "+getVersion());
	  System.err.println("  Developed by Eisenhut Informatik AG, CH-3401 Burgdorf");
	  System.err.println("  See http://www.interlis.ch for information about INTERLIS");
	  System.err.println("  Parts of this program have been generated by ANTLR; see http://www.antlr.org");
	  System.err.println("  This product includes software developed by the");
	  System.err.println("  Apache Software Foundation (http://www.apache.org/).");
	}


	private void printDescription ()
	{
	  System.err.println("DESCRIPTION");
	  System.err.println("  Translates INTERLIS 2 data model definitions to a "+getDB_PRODUCT_NAME()+" schema.");
	  System.err.println("  Loads INTERLIS 2 data into a "+getDB_PRODUCT_NAME()+" database.");
	  System.err.println("  Unloads INTERLIS 2 data from a "+getDB_PRODUCT_NAME()+" database.");
	}


	private void printUsage()
	{
	  System.err.println ("USAGE");
	  System.err.println("  java -jar "+getJAR_NAME()+" [Options] [file.xtf]");
	}
	private  String version=null;
	public String getVersion() {
		  if(version==null){
		java.util.ResourceBundle resVersion = java.util.ResourceBundle.getBundle(ch.ehi.basics.i18n.ResourceBundle.class2qpackageName(this.getClass())+".Version");
		//java.util.ResourceBundle resVersion = java.util.ResourceBundle.getBundle("ch/ehi/ili2db/Version");
			// Major version numbers identify significant functional changes.
			// Minor version numbers identify smaller extensions to the functionality.
			// Micro versions are even finer grained versions.
			StringBuffer ret=new StringBuffer(20);
		ret.append(resVersion.getString("versionMajor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMinor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMicro"));
			ret.append('-');
		ret.append(resVersion.getString("versionDate"));
			version=ret.toString();
		  }
		  return version;
	}
	public String getAPP_HOME()
	{
	  String classpath = System.getProperty("java.class.path");
	  int index = classpath.toLowerCase().indexOf(getJAR_NAME());
	  int start = classpath.lastIndexOf(java.io.File.pathSeparator,index) + 1;
	  if(index > start)
	  {
		  return classpath.substring(start,index - 1);
	  }
	  return null;
	}
}
