/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db.fromxtf;
import java.util.ArrayList;
import java.util.Vector;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.sql.PreparedStatement;
import java.sql.Connection;
import java.sql.Types;

import ch.ehi.basics.logging.EhiLogger;
import ch.ehi.ili2db.converter.*;
import ch.ehi.ili2db.fromili.TransferFromIli;
import ch.ehi.ili2db.mapping.Mapping;

import ch.interlis.ili2c.metamodel.*;

import ch.interlis.iom.IomConstants;
import ch.interlis.iom.IomObject;
import ch.interlis.iom_j.xtf.XtfReader;
import ch.interlis.iox.*;


/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 17.02.2005 $
 */
public class TransferFromXtf {
	private Mapping ili2sqlName=null;
	/** mappings from xml-tags to ili2c-Viewables 
	 */
	private HashMap tag2class=null;
	/** list of seen but unknown types; maintained to prevent duplicate error messages
	 */
	private HashSet unknownTypev=null;
	private TransferDescription td=null;
	private Connection conn=null;
	private java.sql.Timestamp today=null;
	private String dbusr=null;
	private SqlGeometryConverter geomConv=null;
	private int defaultSrsid=0;
	//private int sqlIdGen=1;
	private ch.ehi.ili2db.base.DbIdGen idGen=null;
	private HashMap xtfId2sqlId=new HashMap();
	/** list of not yet processed struct values
	 */
	private ArrayList structQueue=null;
	public TransferFromXtf(Mapping ili2sqlName1,
			TransferDescription td1,
			Connection conn1,
			String dbusr1,
			SqlGeometryConverter geomConv,
			String defaultCrsAuthority,
			String defaultSrcCode){
		ili2sqlName=ili2sqlName1;
		td=td1;
		conn=conn1;
		dbusr=dbusr1;
		if(dbusr==null || dbusr.length()==0){
			dbusr=System.getProperty("user.name");
		}
		this.geomConv=geomConv;
		idGen=new ch.ehi.ili2db.base.DbIdGen(conn,dbusr);
		try{
			defaultSrsid=geomConv.getSrsid(defaultCrsAuthority,defaultSrcCode,conn);
		}catch(UnsupportedOperationException ex){
			EhiLogger.logAdaption("no CRS support by converter; use -1 as default srsid");
			defaultSrsid=-1;
		}catch(ConverterException ex){
			throw new IllegalArgumentException("failed to get srsid for "+defaultCrsAuthority+":"+defaultSrcCode+", "+ex.getLocalizedMessage());
		}
	}
		
	public void doit(XtfReader reader)
	throws IoxException
	{
		today=new java.sql.Timestamp(System.currentTimeMillis());
		tag2class=ch.interlis.ili2c.generator.XSDGenerator.getTagMap(td);
		unknownTypev=new HashSet();
		structQueue=new ArrayList();
		boolean surfaceAsPolyline=true;
		// more baskets?
		IoxEvent event=reader.read();
		while(event!=null){
			if(event instanceof StartBasketEvent){
				StartBasketEvent basket=(StartBasketEvent)event;
				EhiLogger.logState("Basket "+basket.getType()+"(oid "+basket.getBid()+")...");
			}else if(event instanceof ObjectEvent){
				IomObject iomObj=((ObjectEvent)event).getIomObject();
				// translate object
				try{
					writeObject(iomObj,null);
				}catch(ConverterException ex){
					EhiLogger.logError("Object "+iomObj.getobjectoid()+" at (line "+iomObj.getobjectline()+",col "+iomObj.getobjectcol()+")",ex);
				}catch(java.sql.SQLException ex){
					EhiLogger.logError("Object "+iomObj.getobjectoid()+" at (line "+iomObj.getobjectline()+",col "+iomObj.getobjectcol()+")",ex);
				}
				while(!structQueue.isEmpty()){
					StructWrapper struct=(StructWrapper)structQueue.remove(0); // get front
					try{
						writeObject(struct.getStruct(),struct);
					}catch(ConverterException ex){
						EhiLogger.logError("Object "+iomObj.getobjectoid()+"; Struct at (line "+struct.getStruct().getobjectline()+",col "+struct.getStruct().getobjectcol()+")",ex);
					}catch(java.sql.SQLException ex){
						EhiLogger.logError("Object "+iomObj.getobjectoid()+"; Struct at (line "+struct.getStruct().getobjectline()+",col "+struct.getStruct().getobjectcol()+")",ex);
					}
				}
			}
			event=reader.read();
		}
	}
	/** if structEle==null, iomObj is an object. If structEle!=null iomObj is a struct value.
	 */
	private void writeObject(IomObject iomObj,StructWrapper structEle)
		throws java.sql.SQLException,ConverterException
	{
		String tag=iomObj.getobjecttag();
		//EhiLogger.debug("tag "+tag);
		String sqlType=(String)ili2sqlName.mapIliClassName(tag);
		Viewable aclass=(Viewable)tag2class.get(tag);
		if(aclass==null){
			if(!unknownTypev.contains(tag)){
				EhiLogger.logError("unknown type <"+tag+">, line "+Integer.toString(iomObj.getobjectline())+", col "+Integer.toString(iomObj.getobjectcol()));
			}
			return;
		}
				
		 int sqlId;
		 if(structEle==null){
			sqlId=getObjSqlId(iomObj.getobjectoid());
		 }else{
		 	sqlId=newObjSqlId();
		 }
		 while(aclass!=null){
			String sqlname=(String)ili2sqlName.mapIliClassName(aclass.getScopedName(null));
			String insert = getInsertStmt(sqlname,aclass,geomConv);
			PreparedStatement ps = conn.prepareStatement(insert);
			try{
				int valuei=1;
				ps.setInt(valuei, sqlId);
				valuei++;
				if(aclass.getExtending()==null){
					ps.setString(valuei, sqlType);
					valuei++;
					// if struct, add ref to parent
					if(structEle!=null){
						ps.setInt(valuei, structEle.getParentSqlId());
						valuei++;
						ps.setString(valuei, structEle.getParentSqlType());
						valuei++;
						// T_ParentAttr
						ps.setString(valuei, structEle.getParentSqlAttr());
						valuei++;
						// T_Seq
						ps.setInt(valuei, structEle.getStructi());
						valuei++;
					}
				}
			 
				Iterator iter = aclass.getDefinedAttributesAndRoles2();
				while (iter.hasNext()) {
					ViewableTransferElement obj = (ViewableTransferElement)iter.next();
					if (obj.obj instanceof AttributeDef) {
						AttributeDef attr = (AttributeDef) obj.obj;
						if(attr.getExtending()==null){
							 String attrName=attr.getName();
							Type type = attr.getDomain();
							if( (type instanceof TypeAlias) 
								&& TransferFromIli.isBoolean(td,type)) {
									String value=iomObj.getattrvalue(attrName);
									if(value!=null){
										if(value.equals("true")){
											geomConv.setBoolean(ps,valuei,true);
										}else{
											geomConv.setBoolean(ps,valuei,false);
										}
									}else{
										ps.setNull(valuei,Types.BIT);
									}
									valuei++;
							}else{
								type = attr.getDomainResolvingAliases();
							 
								if (type instanceof CompositionType){
									 // enqueue struct values
									 int structc=iomObj.getattrvaluecount(attrName);
									 for(int structi=0;structi<structc;structi++){
									 	IomObject struct=iomObj.getattrobj(attrName,structi);
									 	String sqlAttrName=ili2sqlName.mapIliAttributeDef(attr);
									 	enqueStructValue(sqlId,sqlType,sqlAttrName,struct,structi);
									 }
								}else if (type instanceof PolylineType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,geomConv.fromIomPolyline(value,getSrsid(type),false));
									 }else{
										geomConv.setPolylineNull(ps,valuei);
									 }
									 valuei++;
								 }else if(type instanceof SurfaceType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,geomConv.fromIomSurface(value,getSrsid(type),((SurfaceType)type).getLineAttributeStructure()!=null,false));
									 }else{
										geomConv.setSurfaceNull(ps,valuei);
									 }
									 valuei++;
								 }else if(type instanceof AreaType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,geomConv.fromIomSurface(value,getSrsid(type),((AreaType)type).getLineAttributeStructure()!=null,false));
									 }else{
										geomConv.setSurfaceNull(ps,valuei);
									 }
									 valuei++;
								 }else if(type instanceof CoordType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										boolean is3D=((CoordType)type).getDimensions().length==3;
										ps.setObject(valuei,geomConv.fromIomCoord(value,getSrsid(type),is3D));
									 }else{
										geomConv.setCoordNull(ps,valuei);
									 }
									 valuei++;
								}else if(type instanceof NumericType){
									String value=iomObj.getattrvalue(attrName);
									if(type.isAbstract()){
									}else{
										PrecisionDecimal min=((NumericType)type).getMinimum();
										PrecisionDecimal max=((NumericType)type).getMaximum();
										if(min.getAccuracy()>0){
											if(value!=null){
												try{
													ps.setDouble(valuei, Double.parseDouble(value));
												}catch(java.lang.NumberFormatException ex){
													EhiLogger.logError(ex);
												}
											}else{
												geomConv.setDecimalNull(ps,valuei);
											}
										}else{
											if(value!=null){
												try{
													int val=(int)Math.round(Double.parseDouble(value));
													ps.setInt(valuei, val);
												}catch(java.lang.NumberFormatException ex){
													EhiLogger.logError(ex);
												}
											}else{
												ps.setNull(valuei,Types.INTEGER);
											}
										}
										valuei++;
									}
								}else{
									String value=iomObj.getattrvalue(attrName);
									if(value!=null){
										ps.setString(valuei, value);
									}else{
										ps.setNull(valuei,Types.VARCHAR);
									}
									valuei++;
								}
							}
						}
					}
					if(obj.obj instanceof RoleDef){
						RoleDef role = (RoleDef) obj.obj;
						if(role.getExtending()==null){
							String roleName=role.getName();
							// a role of an embedded association?
							if(obj.embedded){
								AssociationDef roleOwner = (AssociationDef) role.getContainer();
								if(roleOwner.getDerivedFrom()==null){
									// not just a link?
									 IomObject structvalue=iomObj.getattrobj(roleName,0);
									if (roleOwner.getAttributes().hasNext()
										|| roleOwner.getLightweightAssociations().iterator().hasNext()) {
										 // TODO handle attributes of link
									}
									if(structvalue!=null){
										String refoid=structvalue.getobjectrefoid();
										long orderPos=structvalue.getobjectreforderpos();
										if(orderPos!=0){
										   // refoid,orderPos
										   //ret.setStringAttribute(roleName, refoid);
										   //ret.setStringAttribute(roleName+".orderPos", Long.toString(orderPos));
										}else{
										   // refoid
										   //ret.setStringAttribute(roleName, refoid);
										}
									   int refsqlId=getObjSqlId(refoid);
									   ps.setInt(valuei, refsqlId);
									}else{
										ps.setNull(valuei, Types.INTEGER);
									}
									valuei++;
								}
							 }else{
								 IomObject structvalue=iomObj.getattrobj(roleName,0);
								 String refoid=structvalue.getobjectrefoid();
								 long orderPos=structvalue.getobjectreforderpos();
								 if(orderPos!=0){
									// refoid,orderPos
									//ret.setStringAttribute(roleName, refoid);
									//ret.setStringAttribute(roleName+".orderPos", Long.toString(orderPos));
								 }else{
									// refoid
									//ret.setStringAttribute(roleName, refoid);
								 }
								int refsqlId=getObjSqlId(refoid);
								ps.setInt(valuei, refsqlId);
								valuei++;
							 }
						}
					 }
				}
				// T_LastChange
				ps.setTimestamp(valuei, today);
				valuei++;
				// T_CreateDate
				ps.setTimestamp(valuei, today);
				valuei++;
				// T_User
				ps.setString(valuei, dbusr);
				valuei++;
				ps.executeUpdate();
			}finally{
				ps.close();
			}
			aclass=(Viewable)aclass.getExtending();
		 }
	}
	private void enqueStructValue(int parentSqlId,String parentSqlType,String parentSqlAttr,IomObject struct,int structi)
	{
		structQueue.add(new StructWrapper(parentSqlId,parentSqlType,parentSqlAttr,struct,structi));
	}

	/** maps the xtfId to a sqlId.
	 */
	private int getObjSqlId(String xtfId){
		if(xtfId2sqlId.containsKey(xtfId)){
			return ((Integer)xtfId2sqlId.get(xtfId)).intValue();
		}
		int ret=newObjSqlId();
		xtfId2sqlId.put(xtfId,new Integer(ret));
		return ret;
	}
	/** gets a new obj id.
	 */
	int idBlockStart=0;
	int lastLocalId=0;
	private int newObjSqlId(){
		int ret;
		final int BLOCK_SIZE=20;
		if(lastLocalId!=0 && lastLocalId<BLOCK_SIZE){
			lastLocalId++;
			ret=idBlockStart+lastLocalId;
		}else{
			lastLocalId=1;
			ret=idBlockStart=idGen.getLastId("T_Id")+lastLocalId;
			idGen.setLastId(idBlockStart+BLOCK_SIZE,"T_Id");
		}
		return ret;
	}
	/** creates an insert statement for a given viewable.
	 * @param sqlname table name of viewable
	 * @param aclass viewable
	 * @return insert statement
	 */
	private String createInsertStmt(String sqlname,Viewable aclass,SqlGeometryConverter conv){
		StringBuffer ret = new StringBuffer();
		StringBuffer values = new StringBuffer();
		ret.append("INSERT INTO ");
		ret.append(sqlname);
		String sep=" (";
		Iterator iter = aclass.getDefinedAttributesAndRoles2();
		// add T_Id
		ret.append(sep);
		sep=",";
		ret.append("T_Id");
		values.append("?");
		// if root, add type
		if(aclass.getExtending()==null){
			ret.append(sep);
			sep=",";
			ret.append("T_Type");
			values.append(",?");
			// if STRUCTURE, add ref to parent
			if((aclass instanceof Table) && !((Table)aclass).isIdentifiable()){
				ret.append(sep);
				sep=",";
				ret.append("T_ParentId");
				values.append(",?");
				ret.append(sep);
				sep=",";
				ret.append("T_ParentType");
				values.append(",?");
				// attribute name in parent class
				ret.append(sep);
				sep=",";
				ret.append("T_ParentAttr");
				values.append(",?");
				// seqeunce (not null if LIST)
				ret.append(sep);
				sep=",";
				ret.append("T_Seq");
				values.append(",?");
			}
		}
		while (iter.hasNext()) {
		   ViewableTransferElement obj = (ViewableTransferElement)iter.next();
		   if (obj.obj instanceof AttributeDef) {
			   AttributeDef attr = (AttributeDef) obj.obj;
			   if(attr.getExtending()==null){
				Type type = attr.getDomainResolvingAliases();
				String attrName=ili2sqlName.mapIliAttributeDef(attr);
				if (type instanceof CompositionType){
				}else if (type instanceof PolylineType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
						values.append(","+conv.getInsertValueWrapperPolyline("?",getSrsid(type)));
				 }else if(type instanceof SurfaceType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
						values.append(","+conv.getInsertValueWrapperSurface("?",getSrsid(type)));
				 }else if(type instanceof AreaType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
						values.append(","+conv.getInsertValueWrapperSurface("?",getSrsid(type)));
				 }else if(type instanceof CoordType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
						values.append(","+conv.getInsertValueWrapperCoord("?",getSrsid(type)));
				}else{
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
						values.append(",?");
				}
			   }
		   }
		   if(obj.obj instanceof RoleDef){
			   RoleDef role = (RoleDef) obj.obj;
			   if(role.getExtending()==null){
				String roleName=ili2sqlName.mapIliRoleDef(role);
				// a role of an embedded association?
				if(obj.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						 // TODO if(orderPos!=0){
						 ret.append(sep);
						 sep=",";
						 ret.append(roleName);
							values.append(",?");
					}
				 }else{
					 // TODO if(orderPos!=0){
					 ret.append(sep);
					 sep=",";
					 ret.append(roleName);
						values.append(",?");
				 }
			   }
			}
		}
		// stdcols
		ret.append(sep);
		sep=",";
		ret.append("T_LastChange");
		values.append(",?");
		ret.append(sep);
		sep=",";
		ret.append("T_CreateDate");
		values.append(",?");
		ret.append(sep);
		sep=",";
		ret.append("T_User");
		values.append(",?");

		ret.append(") VALUES (");
		ret.append(values);
		ret.append(")");
		return ret.toString();
	}
	private HashMap insertStmts=new HashMap();
	/** gets an insert statement for a given viewable. Creates only a new
	 *  statement if this is not yet seen sqlname.
	 * @param sqlname table name of viewable
	 * @param aclass viewable
	 * @return insert statement
	 */
	private String getInsertStmt(String sqlname,Viewable aclass,SqlGeometryConverter conv){
		if(insertStmts.containsKey(sqlname)){
			return (String)insertStmts.get(sqlname);
		}
		String stmt=createInsertStmt(sqlname,aclass,conv);
		EhiLogger.traceBackendCmd(stmt);
		insertStmts.put(sqlname,stmt);
		return stmt;
	}
	private int getSrsid(Type type){
		return defaultSrsid;
	}
}
