/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db.mapping;

import java.util.HashMap;
import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.iom.IomObject;
import ch.interlis.iom_j.xtf.XtfReader;
import ch.interlis.iox.*;

import ch.ehi.ili2db.mapping.*;
import ch.ehi.ili2db.base.Ili2dbException;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 04.04.2005 $
 */
public class Mapping {
	public static String SQL_T_ILI2DB_CLASSNAME="T_ILI2DB_CLASSNAME";
	private static String SQL_IliName="IliName";
	private static String SQL_SqlName="SqlName";
	/** mapping from a qualified interlis viewable name to a sql table name.
	 * Maintained by addMApping().
	 */
	private HashMap ili2sqlName=new HashMap();
	/** mapping from a sql table name to a qualified interlis viewable name.
	 * Maintained by addMApping().
	 */
	private HashMap sql2iliName=new HashMap();
	private HashMap config=new HashMap();
	public void readConfig(String filename)
	throws Ili2dbException
	{
		java.io.InputStream inputFile=null;
		try{
			inputFile=new java.io.FileInputStream(new java.io.File (filename));
			XtfReader reader=new XtfReader(inputFile);
			IoxEvent event=reader.read();
			while(event!=null){
				if(event instanceof ObjectEvent){
					IomObject obj=((ObjectEvent)event).getIomObject();
					String tag=obj.getobjecttag();
					if(tag.equals("Ili2ora05.MappingConfig.ModelDef")){
						String iliName=obj.getattrvalue("iliName");
						String sqlName=obj.getattrvalue("sqlName");
						ModelDef model=new ModelDef();
						model.setIliName(iliName);
						if(sqlName!=null){
							model.setSqlName(sqlName);
						}
						config.put(iliName,model);
						int topicc=obj.getattrvaluecount("definition");
						for(int topici=0;topici<topicc;topici++){
							handleModelMember(model,obj.getattrobj("definition",topici));
						}
					}else{
						throw new Ili2dbException("("+obj.getobjectline()+","+obj.getobjectcol()+"): unknown tag "+tag);
					}
				}
			}
		}catch(ch.interlis.iox.IoxException ex){
			throw new Ili2dbException(ex);
		}catch(java.io.IOException ex){
			throw new Ili2dbException(ex);
		}finally{
			if(inputFile!=null){
				try{
					inputFile.close();
				}catch(java.io.IOException ex){
					throw new Ili2dbException(ex);
				}
				inputFile=null;
			}
		}
				
	}
	private void handleModelMember(ModelDef model,IomObject topicObj){
		if(topicObj!=null){
			String tag=topicObj.getobjecttag();
			if(tag.equals("Ili2ora05.MappingConfig.TopicDef")){
				String iliName=topicObj.getattrvalue("iliName");
				String sqlName=topicObj.getattrvalue("sqlName");
				TopicDef topic=new TopicDef();
				topic.setIliName(iliName);
				if(sqlName!=null){
					topic.setSqlName(sqlName);
				}
				model.addDefinition(topic);
				int classc=topicObj.getattrvaluecount("definition");
				for(int memberi=0;memberi<classc;memberi++){
					handleTopicMember(topic,topicObj.getattrobj("definition",memberi));
				}
			}else if(tag.equals("Ili2ora05.MappingConfig.ClassDef")){
				String iliName=topicObj.getattrvalue("iliName");
				String sqlName=topicObj.getattrvalue("sqlName");
				ClassDef aclass=new ClassDef();
				aclass.setIliName(iliName);
				if(sqlName!=null){
					aclass.setSqlName(sqlName);
				}
				model.addDefinition(aclass);
				int attrc=topicObj.getattrvaluecount("attribute");
				for(int memberi=0;memberi<attrc;memberi++){
					handleAttribute(aclass,topicObj.getattrobj("attribute",memberi));
				}
			}else{
				EhiLogger.logError("("+topicObj.getobjectline()+","+topicObj.getobjectcol()+"): unknown tag "+tag);
			}
		}
	}
	private void handleTopicMember(TopicDef topic,IomObject classObj){
		if(classObj!=null){
			String tag=classObj.getobjecttag();
			if(tag.equals("Ili2ora05.MappingConfig.ClassDef")){
				String iliName=classObj.getattrvalue("iliName");
				String sqlName=classObj.getattrvalue("sqlName");
				ClassDef aclass=new ClassDef();
				aclass.setIliName(iliName);
				if(sqlName!=null){
					aclass.setSqlName(sqlName);
				}
				topic.addDefinition(aclass);
				int attrc=classObj.getattrvaluecount("attribute");
				for(int memberi=0;memberi<attrc;memberi++){
					handleAttribute(aclass,classObj.getattrobj("attribute",memberi));
				}
			}else{
				EhiLogger.logError("("+classObj.getobjectline()+","+classObj.getobjectcol()+"): unknown tag "+tag);
			}
		}
	}
	private void handleAttribute(ClassDef aclass,IomObject attrObj){
		if(attrObj!=null){
			String tag=attrObj.getobjecttag();
			if(tag.equals("Ili2ora05.MappingConfig.AttributeDef")){
				String iliName=attrObj.getattrvalue("iliName");
				String sqlName=attrObj.getattrvalue("sqlName");
				AttributeDef attr=new AttributeDef();
				attr.setIliName(iliName);
				if(sqlName!=null){
					attr.setSqlName(sqlName);
				}
				aclass.addAttribute(attr);
			}else{
				EhiLogger.logError("("+attrObj.getobjectline()+","+attrObj.getobjectcol()+"): unknown tag "+tag);
			}
		}
	}
	private String map(String modelName,String topicName,String className)
	{
		StringBuffer ret=new StringBuffer();
		String modelSqlName=modelName;
		String topicSqlName=topicName;
		String classSqlName=className;
		ModelDef model=(ModelDef)config.get(modelName);
		if(model!=null){
			String sqlName=model.getSqlName();
			if(sqlName!=null){
				modelSqlName=sqlName;
			}
			if(topicName!=null){
				TopicDef topic=(TopicDef)model.getDefinition(topicName);
				if(topic!=null){
					sqlName=topic.getSqlName();
					if(sqlName!=null){
						topicSqlName=sqlName;
					}
					ClassDef aclass=(ClassDef)topic.getDefinition(className);
					if(aclass!=null){
						sqlName=aclass.getSqlName();
						if(sqlName!=null){
							classSqlName=sqlName;
						}
					}
				}
			}else{
				ClassDef aclass=(ClassDef)model.getDefinition(className);
				if(aclass!=null){
					sqlName=aclass.getSqlName();
					if(sqlName!=null){
						classSqlName=sqlName;
					}
				}
			}
		}
		if(topicSqlName==null){
			ret.append(shortcutName(modelSqlName,13));
		}else{
			ret.append(shortcutName(modelSqlName,5));
			ret.append(shortcutName(topicSqlName,8));
		}
		ret.append("_");
		ret.append(shortcutName(classSqlName,30-ret.length()));
		return ret.toString();
	}
	public String mapSqlTableName(String sqlname){
		return (String)sql2iliName.get(sqlname);
	}
	/** maps a qualified interlis class name to a sql table name.
	 * @param iliname qualified interlis table name
	 * @return sql table name
	 */
	public String mapIliClassName(String iliname)
	{
		String sqlname=(String)ili2sqlName.get(iliname);
		if(sqlname==null){
			int pos1=iliname.indexOf('.');
			int pos2=iliname.indexOf('.',pos1+1);
			String modelName=iliname.substring(0,pos1);
			String topicName=null;
			String className=null;
			if(pos2!=-1){
				topicName=iliname.substring(pos1+1,pos2);
				className=iliname.substring(pos2+1);
			}else{
				topicName=null;
				className=iliname.substring(pos1+1);
			}
			sqlname=map(modelName,topicName,className);
			addMapping(iliname,sqlname);
		}
		return sqlname;
	}
	private void addMapping(String iliname,String sqlname)
	{
		ili2sqlName.put(iliname,sqlname);
		sql2iliName.put(sqlname,iliname);
	}
	public String mapIliClassDef(ch.interlis.ili2c.metamodel.Viewable def)
	{
		String iliname=def.getScopedName(null);
		String sqlname=(String)ili2sqlName.get(iliname);
		if(sqlname==null){
			ch.interlis.ili2c.metamodel.Topic topic=(ch.interlis.ili2c.metamodel.Topic)def.getContainer(ch.interlis.ili2c.metamodel.Topic.class);
			ch.interlis.ili2c.metamodel.Model model=(ch.interlis.ili2c.metamodel.Model)def.getContainer(ch.interlis.ili2c.metamodel.Model.class);
			sqlname=map(model.getName(),topic!=null ? topic.getName():null,def.getName());
			addMapping(iliname,sqlname);
		}
		return sqlname;
	}
	public String mapIliRoleDef(ch.interlis.ili2c.metamodel.RoleDef def){
		String ret= shortcutName(def.getName(),30);
		return makeValidSqlName(ret);
	}
	public String mapIliAttributeDef(ch.interlis.ili2c.metamodel.AttributeDef def){
		String ret=shortcutName(def.getName(),30);
		return makeValidSqlName(ret);
	}
	private String makeValidSqlName(String name)
	{
		if(name.equalsIgnoreCase("Date")){
		  return "adate";
		}
		if(name.equalsIgnoreCase("level")){
		  return "alevel";
		}
		if(name.equalsIgnoreCase("number")){
		  return "anumber";
		}
		// MS-ACCESS
		if(name.equalsIgnoreCase("value")){
		  return "avalue";
		}
		return name;
	}
	private static String shortcutName(String aname, int maxlen)
	{
		StringBuffer name=new StringBuffer(aname);
		// number of charcters to remove
	int stripc=name.length()-maxlen;
	if(stripc<=0)return name.toString();
		// remove vocals
	for(int i=name.length()-4;i>=3;i--){
		char c=name.charAt(i);
		if(c=='a' || c=='e' || c=='i' || c=='o' || c=='u'
			|| c=='A' || c=='E' || c=='I' || c=='O' || c=='U'){
			name.deleteCharAt(i);
			stripc--;
			if(stripc==0)return name.toString();
		}
	}
		// still to long
		// remove from the middle of the name
		int start=(name.length()-stripc)/2;
	name.delete(start,start+stripc);
	// ASSERT(!name.IsEmpty());
	return name.toString();
	}
	static public void addMappingTable(ch.ehi.sqlgen.repository.DbSchema schema)
	{
		ch.ehi.sqlgen.repository.DbTable tab=new ch.ehi.sqlgen.repository.DbTable();
		tab.setName(SQL_T_ILI2DB_CLASSNAME);
		ch.ehi.sqlgen.repository.DbColVarchar thisClass=new ch.ehi.sqlgen.repository.DbColVarchar();
		thisClass.setName(SQL_IliName);
		thisClass.setNotNull(true);
		thisClass.setSize(1024);
		tab.addColumn(thisClass);
		ch.ehi.sqlgen.repository.DbColVarchar baseClass=new ch.ehi.sqlgen.repository.DbColVarchar();
		baseClass.setName(SQL_SqlName);
		baseClass.setNotNull(true);
		baseClass.setSize(30);
		tab.addColumn(baseClass);
		schema.addTable(tab);
	}

	public void updateMappingTable(java.sql.Connection conn)
	throws Ili2dbException
	{
		// create table
		try{

			// insert mapping entries
			String stmt="INSERT INTO "+SQL_T_ILI2DB_CLASSNAME+" ("+SQL_IliName+","+SQL_SqlName+") VALUES (?,?)";
			java.sql.PreparedStatement ps = conn.prepareStatement(stmt);
			String iliname=null;
			String sqlname=null;
			try{
				java.util.Iterator entri=ili2sqlName.keySet().iterator();
				while(entri.hasNext()){
					iliname=(String)entri.next();
					sqlname=(String)ili2sqlName.get(iliname);
					ps.setString(1, iliname);
					ps.setString(2, sqlname);
					ps.executeUpdate();
				}
			}catch(java.sql.SQLException ex){
				throw new Ili2dbException("failed to insert classname-mapping "+iliname,ex);
			}finally{
				ps.close();
			}
		}catch(java.sql.SQLException ex){		
			throw new Ili2dbException("failed to update mapping-table "+SQL_T_ILI2DB_CLASSNAME,ex);
		}

	}
	public void readMappingTable(java.sql.Connection conn)
	throws Ili2dbException
	{
		// create table
		String stmt="SELECT "+SQL_IliName+", "+SQL_SqlName+" FROM "+SQL_T_ILI2DB_CLASSNAME;
		java.sql.Statement dbstmt = null;
		try{
			
			dbstmt = conn.createStatement();
			java.sql.ResultSet rs=dbstmt.executeQuery(stmt);
			while(rs.next()){
				String iliname=rs.getString(SQL_IliName);
				String sqlname=rs.getString(SQL_SqlName);
				//EhiLogger.debug("map: "+iliname+"->"+sqlname);
				addMapping(iliname,sqlname);
			}
		}catch(java.sql.SQLException ex){		
			throw new Ili2dbException("failed to query mapping-table "+SQL_T_ILI2DB_CLASSNAME,ex);
		}finally{
			if(dbstmt!=null){
				try{
					dbstmt.close();
				}catch(java.sql.SQLException ex){		
					throw new Ili2dbException("failed to close query of "+SQL_T_ILI2DB_CLASSNAME,ex);
				}
			}
		}

	}
}
